@@ -6,6 +6,7 @@ use Carmel;
 use Carmel::Runner;
 use Carp ();
 use Carmel::Builder;
+use Carmel::CPANfile;
 use Carmel::Repository;
 use Carmel::Resolver;
 use Config qw(%Config);
@@ -193,42 +194,73 @@ sub update_dependencies {
     $self->save_snapshot(\@artifacts);
 }
 
-sub install {
-    my($self, $root_reqs, $snapshot) = @_;
-
-    my $requirements = CPAN::Meta::Requirements->new;
-
-    $self->resolver(
-        root => $root_reqs,
-        snapshot => $snapshot,
-        found => sub {
-            my($artifact) = @_;
-            printf "Using %s (%s)\n", $artifact->package, $artifact->version || '0';
-        },
-        missing => sub {
-            my($module, $want_version, $dist) = @_;
-            $requirements->add_string_requirement($module => $want_version);
-        },
-    )->resolve;
-
-    if (my @missing = $requirements->required_modules) {
-        my $cpanfile = Module::CPANfile->from_prereqs({
-            runtime => {
-                requires => $requirements->as_string_hash,
-            },
-        });
-        print "---> Installing new dependencies: ", join(", ", @missing), "\n";
-        my $builder = $self->builder(cpanfile => $cpanfile, snapshot => $snapshot);
-        $builder->install;
-    }
+sub resolve_dependencies {
+    my($self, $root_reqs, $missing, $snapshot) = @_;
 
     my @artifacts;
     $self->resolver(
-        found    => sub { push @artifacts, $_[0] },
         root     => $root_reqs,
         snapshot => $snapshot,
+        found    => sub {
+            my $artifact = shift;
+            printf "Using %s (%s)\n", $artifact->package, $artifact->version || '0';
+            push @artifacts, $artifact;
+        },
+        missing  => sub {
+            my($module, $want_version) = @_;
+            $missing->add_string_requirement($module => $want_version);
+        },
     )->resolve;
 
+    return @artifacts;
+}
+
+sub is_identical_requirement {
+    my($self, $old, $new) = @_;
+
+    return unless $old;
+
+    # not super accurate but enough
+    join(',', sort $old->required_modules) eq join(',', sort $new->required_modules);
+}
+
+sub try_install {
+    my($self, $root_reqs, $snapshot) = @_;
+
+    my $prev;
+    while (1) {
+        my $missing = CPAN::Meta::Requirements->new;
+        my @artifacts = $self->resolve_dependencies($root_reqs, $missing, $snapshot);
+
+        if (!$missing->required_modules) {
+            return @artifacts;
+        }
+
+        if ($self->is_identical_requirement($prev, $missing)) {
+            my $prereqs = $missing->as_string_hash;
+            my $requirements = join ", ", map "$_ => @{[ $prereqs->{$_} || '0' ]}", keys %$prereqs;
+            die "Can't find an artifact for $requirements\n" .
+              "You need to run `carmel install` first to get the modules installed and artifacts built.\n";
+        }
+
+        $prev = $missing;
+
+        my $cpanfile = Module::CPANfile->from_prereqs({
+            runtime => {
+                requires => $missing->as_string_hash,
+            },
+        });
+        print "---> Installing new dependencies: ", join(", ", $missing->required_modules), "\n";
+        my $builder = $self->builder(cpanfile => $cpanfile, snapshot => $snapshot);
+        $builder->install;
+    }
+}
+
+sub install {
+    my($self, $root_reqs, $snapshot) = @_;
+
+    my @artifacts = $self->try_install($root_reqs, $snapshot);
+
     # $root_reqs has been mutated at this point. Reload requirements
     printf "---> Complete! %d cpanfile dependencies. %d modules installed.\n",
       scalar(grep { $_ ne 'perl' } $self->requirements->required_modules), scalar(@artifacts);
@@ -241,7 +273,7 @@ sub builder {
 
     Carmel::Builder->new(
         repository_base => $self->repository_base,
-        cpanfile_path => scalar $self->try_cpanfile,
+        cpanfile_path => $self->cpanfile_path,
         collect_artifact => sub { $self->repo->import_artifact(@_) },
         @args,
     );
@@ -281,8 +313,7 @@ sub save_snapshot {
     require Carton::Snapshot;
     require Carton::Dist;
 
-    my $cpanfile = $self->try_cpanfile;
-    my $snapshot = Carton::Snapshot->new(path => $cpanfile . ".snapshot");
+    my $snapshot = Carton::Snapshot->new(path => $self->cpanfile->snapshot_path);
 
     for my $artifact (@$artifacts) {
         my $dist = Carton::Dist->new(
@@ -316,10 +347,7 @@ sub dump_bootstrap {
         %modules = (%modules, $artifact->module_files);
     }
 
-    my $cpanfile = $self->try_cpanfile
-      or die "Can't locate 'cpanfile' to load module list.\n";
-
-    my $prereqs = Module::CPANfile->load($cpanfile)->prereqs->as_string_hash;
+    my $prereqs = $self->cpanfile->load->prereqs->as_string_hash;
     my $package = "Carmel::MySetup"; # hide from PAUSE
 
     my $file = Path::Tiny->new(".carmel/MySetup.pm");
@@ -409,7 +437,7 @@ sub resolver {
 }
 
 sub missing_default {
-    my($self, $module, $want_version, $dist, $depth) = @_;
+    my($self, $module, $want_version, $depth) = @_;
     die "Can't find an artifact for $module => $want_version\n" .
       "You need to run `carmel install` first to get the modules installed and artifacts built.\n";
 }
@@ -526,16 +554,17 @@ sub build_index {
     $index;
 }
 
-sub try_cpanfile {
+sub cpanfile_path {
     my $self = shift;
-    $self->locate_cpanfile($ENV{PERL_CARMEL_CPANFILE});
+    $self->{cpanfile_path} ||= Path::Tiny->new($self->locate_cpanfile)->absolute;
 }
 
 sub locate_cpanfile {
-    my($self, $path) = @_;
+    my $self = shift;
 
+    my $path = $ENV{PERL_CARMEL_CPANFILE};
     if ($path) {
-        return Path::Tiny->new($path)->absolute;
+        return $path;
     }
 
     my $current  = Path::Tiny->cwd;
@@ -543,35 +572,28 @@ sub locate_cpanfile {
 
     until ($current eq '/' or $current eq $previous) {
         my $try = $current->child('cpanfile');
-        return $try->absolute if $try->is_file;
+        return $try if $try->is_file;
         ($previous, $current) = ($current, $current->parent);
     }
 
-    return;
+    return 'cpanfile'; # fallback, most certainly fails later
+}
+
+sub cpanfile {
+    my $self = shift;
+    Carmel::CPANfile->new(path => $self->cpanfile_path);
 }
 
 sub requirements {
     my $self = shift;
 
-    my $cpanfile = $self->try_cpanfile
-      or die "Can't locate 'cpanfile' to load module list.\n";
-
-    return Module::CPANfile->load($cpanfile)
-      ->prereqs->merged_requirements(['runtime', 'test', 'develop'], ['requires']);
+    return $self->cpanfile->load->prereqs
+      ->merged_requirements(['runtime', 'test', 'develop'], ['requires']);
 }
 
 sub snapshot {
     my $self = shift;
-
-    my $cpanfile = $self->try_cpanfile;
-    if ($cpanfile && -e "$cpanfile.snapshot") {
-        require Carton::Snapshot;
-        my $snapshot = Carton::Snapshot->new(path => "$cpanfile.snapshot");
-        $snapshot->load;
-        return $snapshot;
-    }
-
-    return;
+    $self->cpanfile->load_snapshot;
 }
 
 1;
